module PackageCloud
  module CLI
    class Entry < Base
      desc "repository SUBCMD ...ARGS", "manage repositories"
      subcommand "repository", Repository

      desc "distro SUBCMD ...ARGS", "manage repositories"
      subcommand "distro", Distro

      desc "master_token SUBCMD ...ARGS", "manage master tokens"
      subcommand "master_token", MasterToken

      desc "read_token SUBCMD ...ARGS", "manage read tokens"
      subcommand "read_token", ReadToken

      desc "yank user/repo[/distro/version] package_name",
           "yank package from user/repo [in dist/version]"
      def yank(repo_desc, package_name)
        ARGV.clear # otherwise gets explodes

        # strip os/dist
        repo_name = repo_desc.split("/")[0..1].join("/")
        dist = repo_desc.split("/")[2..3].join("/")

        if dist == "" && package_name =~ /\.gem$/
          dist = "gems"
        end

        print "Looking for repository at #{repo_name}... "
        repo = client.repository(repo_desc)
        print "success!\n"

        print "Attempting to yank package at #{repo_name}/#{dist}/#{package_name}..."
        packages = repo.yank(dist, package_name)
        puts "done!".color(:green)
      end

      desc "push user/repo[/distro/version] /path/to/packages",
           "push package(s) to repository (in distro/version if required)"
      option "skip-file-ext-validation", :type => :boolean
      option "yes", :type => :boolean
      option "skip-errors", :type => :boolean
      def push(repo, package_file, *package_files)
        total_time = Benchmark.measure do
          ARGV.clear # otherwise gets explodes
          package_files << package_file

          exts = package_files.map { |f| f.split(".").last }.uniq

          if package_files.length > 1 && exts.length > 1
            abort("You can't push multiple packages of different types at the same time.\nFor example, use *.deb to push all your debs at once.".color(:red))
          end

          invalid_packages = package_files.select do |f|
            !["gem", "deb", "rpm", "dsc", "gz", "bz2", "z", "tar", "egg-info", "zip", "whl", "egg", "jar"].include?(f.split(".").last.downcase)
          end

          if !options.has_key?("skip-file-ext-validation") && invalid_packages.any?
            message = "I don't know how to push these packages:\n\n".color(:red)
            invalid_packages.each do |p|
              message << "  #{p}\n"
            end
            message << "\npackage_cloud only supports debs, gems, jars, python packages, and rpms.".color(:red)
            abort(message)
          end

          if !options.has_key?("yes") && exts.first == "gem" && package_files.length > 1
            answer = get_valid("Are you sure you want to push #{package_files.length} packages? (y/n)") do |s|
              s == "y" || s == "n"
            end

            if answer != "y"
              abort("Aborting...".color(:red))
            end
          end

          validator = Validator.new(client)
          if ["gz", "bz2", "z", "tar", "egg-info", "zip", "whl", "egg"].include?(exts.first.downcase)
            dist_id   = validator.distribution_id(repo, package_files, 'py')
          elsif ["jar"].include?(exts.first.downcase)
            abort_if_snapshot!(package_files)
            dist_id   = validator.distribution_id(repo, package_files, 'jar')
          else
            dist_id   = validator.distribution_id(repo, package_files, exts.first)
          end

          # strip os/dist
          split_repo = repo.split("/")[0..1].join("/")

          print "Looking for repository at #{split_repo}... "
          client_repo = nil
          measurement = Benchmark.measure do
            client_repo = client.repository(split_repo)
          end
          print "success!\n"
          $logger.debug("repository lookup request timing: #{measurement}")

          package_files.each do |f|
            files = nil
            ext = f.split(".").last

            if ext == "dsc"
              print "Checking source package #{f}... "
              files = parse_and_verify_dsc(client_repo, f, dist_id)
            end

            print "Pushing #{f}... "
            measurement = Benchmark.measure do
              if options.has_key?("skip-errors")
                create_package_skip_errors(client_repo, f, dist_id, files, ext)
              else
                create_package(client_repo, f, dist_id, files, ext)
              end
            end
            $logger.debug("create package request timing: #{measurement}")
          end
        end
        $logger.debug("push command total timing: #{total_time}")
      end

      desc "version",
           "print version information"
      def version
        puts "package_cloud CLI #{VERSION}\nSee https://packagecloud.io/docs#cli for more details."
      end

      private
        def abort_if_snapshot!(files)
          if files.any? { |file| file.include?("-SNAPSHOT")  }
            abort("SNAPSHOT uploads are not supported by the CLI, please use Maven instead: https://packagecloud.io/docs#wagon-instructions")
          end
        end

        def create_package_skip_errors(client_repo, f, dist_id, files, ext)
          begin
            client_repo.create_package(f, dist_id, files, ext)
          rescue RestClient::UnprocessableEntity => e
            print "error (skipping):\n".color(:yellow)
            json = JSON.parse(e.response)
            json.each do |k,v|
              puts "\n\t#{k}: #{v.join(", ")}\n"
            end
            puts ""
          end
        end

        def create_package(client_repo, f, dist_id, files, ext)
          begin
            client_repo.create_package(f, dist_id, files, ext)
          rescue RestClient::UnprocessableEntity => e
            print "error:\n".color(:red)
            json = JSON.parse(e.response)
            json.each do |k,v|
              if v.is_a? String
                puts "\n\t#{k}: #{v}\n"
              elsif v.is_a? Array
                puts "\n\t#{k}: #{v.join(", ")}\n"
              end
            end
            puts ""
            exit(1)
          end
        end

        def parse_and_verify_dsc(repo, f, dist_id)
          files = repo.parse_dsc(f, dist_id)
          dirname = File.dirname(f)
          find_and_verify(dirname, files)
        end

        def find_and_verify(dir, files)
          file_paths = []
          files.each do |f|
            filepath = File.join(dir, f["filename"])
            if !File.exists?(filepath)
              print "Unable to find file name: #{f["filename"]} for source package: #{filepath}\n".color(:red)
              abort("Aborting...".color(:red))
            end

            disk_size = File.stat(filepath).size
            if disk_size != f["size"]
              print "File #{f["filename"]} has size: #{disk_size}, expected: #{f["size"]}\n".color(:red)
              abort("Aborting...".color(:red))
            end

            print "Found DSC package file #{f["filename"]} for upload\n"

            file_paths << filepath
          end
          file_paths
        end

        def get_valid(prompt)
          selection = ""
          times = 0
          until yield(selection)
            if times > 0
              puts "#{selection} is not a valid selection."
            end
            print "#{prompt}: "
            selection = ::Kernel.gets.chomp
            times += 1
          end

          selection
        end
    end
  end
end
